@@ -42,20 +42,18 @@ namespace ipe {
 
   class Bitmap {
   public:
-    enum TFilter { EDirect, EFlateDecode, EDCTDecode };
-    enum TColorSpace { EDeviceRGB, EDeviceGray, EDeviceCMYK };
-
-    class MRenderData {
-    public:
-      virtual ~MRenderData();
+    enum Flags {
+      ERGB = 0x01,      // not grayscale
+      EAlpha = 0x02,    // has alpha channel
+      EDCT = 0x04,      // DCT encoded jpg image
+      EInflate = 0x08,  // data needs to be inflated
+      ENative = 0x10,   // data is already in native-endian ARGB32
     };
 
     Bitmap();
-    Bitmap(int width, int height,
-	   TColorSpace colorSpace, int bitsPerComponent,
-	   Buffer data, TFilter filter, bool deflate = false);
+    Bitmap(int width, int height, uint32_t flags, Buffer data);
     Bitmap(const XmlAttributes &attr, String data);
-    Bitmap(const XmlAttributes &attr, Buffer data);
+    Bitmap(const XmlAttributes &attr, Buffer data, Buffer smask);
 
     Bitmap(const Bitmap &rhs);
     ~Bitmap();
@@ -66,59 +64,48 @@ namespace ipe {
     inline bool isNull() const;
     bool equal(Bitmap rhs) const;
 
-    inline TColorSpace colorSpace() const;
-    inline TFilter filter() const;
-    inline int components() const;
-    inline int bitsPerComponent() const;
     inline int width() const;
     inline int height() const;
 
-    int colorKey() const;
-    void setColorKey(int key);
+    inline bool isJpeg() const;
+    inline bool isGray() const;
+    inline bool hasAlpha() const;
+    inline int colorKey() const;
 
-    inline const char *data() const;
-    inline int size() const;
+    Buffer pixelData();
 
     inline int objNum() const;
     inline void setObjNum(int objNum) const;
 
-    inline MRenderData *renderData() const;
-    void setRenderData(MRenderData *data) const;
-
-    Buffer pixelData() const;
+    std::pair<Buffer, Buffer> embed() const;
 
     inline bool operator==(const Bitmap &rhs) const;
     inline bool operator!=(const Bitmap &rhs) const;
     inline bool operator<(const Bitmap &rhs) const;
 
     static const char *readJpegInfo(FILE *file, int &width, int &height,
-				    Vector &dotsPerInch,
-				    TColorSpace &colorSpace,
-				    int &bitsPerComponent);
-
-    static Bitmap readJpeg(const char *fname, Vector &dotsPerInch,
-			   const char * &errmsg);
-    static Bitmap readPNG(const char *fname, bool deflate,
-			  Vector &dotsPerInch, const char * &errmsg);
+				    Vector &dotsPerInch, uint32_t &flags);
+    static Bitmap readJpeg(const char *fname, Vector &dotsPerInch, const char * &errmsg);
+    static Bitmap readPNG(const char *fname, Vector &dotsPerInch, const char * &errmsg);
 
   private:
-    int init(const XmlAttributes &attr);
+    std::pair<int, int> init(const XmlAttributes &attr);
     void computeChecksum();
+    void unpack(Buffer alphaChannel);
+    void analyze();
 
   private:
     struct Imp {
       int iRefCount;
-      TColorSpace iColorSpace;
-      int iBitsPerComponent;
+      uint32_t iFlags;
       int iWidth;
       int iHeight;
-      int iComponents;
       int iColorKey;
-      Buffer iData;
-      TFilter iFilter;
+      Buffer iData;               // native-endian ARGB32 or DCT encoded
+      Buffer iPixelData;          // native-endian ARGB32 pre-multiplied for Cairo
+      bool iPixelsComputed;
       int iChecksum;
-      mutable int iObjNum;           // Object number (e.g. in PDF file)
-      mutable MRenderData *iRender;  // cached pixmap to render it fast
+      mutable int iObjNum;        // Object number (e.g. in PDF file)
     };
 
     Imp *iImp;
@@ -129,25 +116,7 @@ namespace ipe {
   //! Is this a null bitmap?
   inline bool Bitmap::isNull() const
   {
-    return (iImp == 0);
-  }
-
-  //! Return the color space of the image.
-  inline Bitmap::TColorSpace Bitmap::colorSpace() const
-  {
-    return iImp->iColorSpace;
-  }
-
-  //! Return number of components per pixel.
-  inline int Bitmap::components() const
-  {
-    return iImp->iComponents;
-  }
-
-  //! Return the number of bits per component.
-  inline int Bitmap::bitsPerComponent() const
-  {
-    return iImp->iBitsPerComponent;
+    return (iImp == nullptr);
   }
 
   //! Return width of pixel array.
@@ -162,22 +131,29 @@ namespace ipe {
     return iImp->iHeight;
   }
 
-  //! Return the data filter of the image data.
-  inline Bitmap::TFilter Bitmap::filter() const
+  //! Is this bitmap a JPEG photo?
+  inline bool Bitmap::isJpeg() const
   {
-    return iImp->iFilter;
+    return (iImp->iFlags & EDCT) != 0;
   }
 
-  //! Return a pointer to the image data (in PDF arrangement).
-  inline const char *Bitmap::data() const
+  //! Is the bitmap grayscale?
+  inline bool Bitmap::isGray() const
   {
-    return iImp->iData.data();
+    return (iImp->iFlags & ERGB) == 0;
   }
 
-  //! Return size (number of bytes) of image data (in PDF arrangement).
-  inline int Bitmap::size() const
+  //! Does the bitmap have transparency?
+  /*! Bitmaps with color key will return false here. */
+  inline bool Bitmap::hasAlpha() const
   {
-    return iImp->iData.size();
+    return (iImp->iFlags & EAlpha) != 0;
+  }
+
+  //! Return the color key or -1 if none.
+  inline int Bitmap::colorKey() const
+  {
+    return iImp->iColorKey;
   }
 
   //! Return object number of the bitmap.
@@ -192,12 +168,6 @@ namespace ipe {
     iImp->iObjNum = objNum;
   }
 
-  //! Return cached bitmap for rendering.
-  inline Bitmap::MRenderData *Bitmap::renderData() const
-  {
-    return iImp->iRender;
-  }
-
   //! Two bitmaps are equal if they share the same data.
   inline bool Bitmap::operator==(const Bitmap &rhs) const
   {
