@@ -37,6 +37,10 @@
 #include "ipedoc.h"
 #include "ipeimage.h"
 
+#include <list>
+#include <map>
+#include <unordered_map>
+
 // --------------------------------------------------------------------
 
 namespace ipe {
@@ -83,19 +87,20 @@ namespace ipe {
 
   class PdfWriter {
   public:
-    PdfWriter(TellStream &stream, const Document *doc,
-	      const PdfResources *resources,
-	      bool markedView, int fromPage, int toPage, int compression);
+    PdfWriter(TellStream &stream, const Document *doc, const PdfResources *resources,
+	      uint32_t flags, int fromPage, int toPage, int compression);
     ~PdfWriter();
 
     void createPages();
     void createPageView(int page, int view);
     void createBookmarks();
+    void createNamedDests();
     void createXmlStream(String xmldata, bool preCompressed);
     void createTrailer();
 
   private:
     int startObject(int objnum = -1);
+    int pageObjectNumber(int page, int view);
     void createStream(const char *data, int size, bool preCompressed);
     void writeString(String text);
     void embedBitmap(Bitmap bitmap);
@@ -113,12 +118,14 @@ namespace ipe {
     TellStream &iStream;
     const Document *iDoc;
     const PdfResources *iResources;
-    //! Show only last view of each page?
-    bool iMarkedView;
+    //! SaveFlag's
+    uint32_t iSaveFlags;
     //! Obj id of XML stream.
     int iXmlStreamNum;
     //! Obj id of outline dictionary.
     int iBookmarks;
+    //! Obj id of named destinations.
+    int iDests;
     //! Compression level (0..9).
     int iCompressLevel;
     //! Obj id of GState with opacity definitions.
@@ -140,8 +147,13 @@ namespace ipe {
     std::map<int, int> iGradients;
     //! Object numbers of symbols, indexed by attribute name
     std::map<int, int> iSymbols;
+    struct PON {
+      int page;
+      int view;
+      int objNum;
+    };
     //! List of pages, expressed as Pdf object numbers.
-    std::vector<int> iPageObjectNumbers;
+    std::vector<PON> iPageObjectNumbers;
     //! List of file locations, in object number order (starting with 0).
     std::map<int, long> iXref;
   };
