@@ -4,7 +4,7 @@
 --[[
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2012  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -37,7 +37,7 @@ function MODEL:action(a)
   local result, err = xpcall(function () self:paction(a) end,
 			     debug.traceback)
   if not result then
-    ipeui.messageBox(nil, "critical",
+    messageBox(nil, "critical",
 		     "Lua error\n\n"..
 		       "Data may have been corrupted. \n" ..
 		       "Save your file!",
@@ -45,16 +45,36 @@ function MODEL:action(a)
   end
 end
 
-function MODEL:paction(a)
+function MODEL:paction(a1)
+  -- work-around for bug in Qt 5.5, to be replaced with something better:
+  local a = a1:gsub("&", "")
+  -- print("MODEL:paction(" .. a .. ")")
   if a:sub(1,5) == "mode_" then
     self.mode = a:sub(6)
-    if prefs.tablet_mode then
-      self.ui:setSelectionVisible(self.mode ~= "ink")
-      self.ui:setCursor()
+    self.ui:setInkMode(self.mode == "ink")
+    if prefs.no_ink_snap then
+      -- disable snapping in ink mode
+      if self.mode == "ink" then
+	self.savedsnap = {}
+	for i,e in ipairs(MODEL.snapmodes) do
+	  self.savedsnap[e] = self.snap[e]
+	  self.snap[e] = false
+	end
+      elseif self.savedsnap then
+	-- reenable snapping when switching back to another mode
+	for i,e in ipairs(MODEL.snapmodes) do
+	  self.snap[e] = self.savedsnap[e]
+	end
+	self.savedsnap = nil
+      end
+      for i,e in ipairs(MODEL.snapmodes) do
+	self.ui:setActionState(e, self.snap[e])
+      end
+      self:setSnap()
     end
   elseif a:sub(1,4) == "snap" then
     self.snap[a] = self.ui:actionState(a)
-    self.ui:setSnap(self.snap)
+    self:setSnap()
     self.ui:setFifiVisible(true)
   elseif a:sub(1,7) == "ipelet_" then
     self:action_ipelet(a)
@@ -79,13 +99,21 @@ function MODEL:paction(a)
   end
 end
 
+local function has_text(page)
+  for i,obj,sel,layer in page:objects() do
+    if sel and obj:type() == "text" then return true end
+  end
+  return false
+end
+
 -- Attribute selector
 function MODEL:selector(prop, value)
   if prop == "gridsize" or prop == "anglesize" then
     local abs = self.doc:sheets():find(prop, value)
     self.snap[prop] = abs
-    self.ui:setSnap(self.snap)
+    self:setSnap()
     self:setPage()
+    self.ui:setGridAngleSize(self.snap.gridsize, self.snap.anglesize)
     return
   end
   if prop == "markshape" then
@@ -104,6 +132,9 @@ function MODEL:selector(prop, value)
   self.ui:setAttributes(self.doc:sheets(), self.attributes)
   if self:page():hasSelection() then
     self:setAttribute(prop, value)
+    if prop == "textsize" or has_text(self:page()) then
+      self:autoRunLatex()
+    end
   end
   -- self:print_attributes()
 end
@@ -141,30 +172,39 @@ function MODEL:bookmark(index)
   self:setPage()
 end
 
+function MODEL:findOldValue(button, btype)
+  local old = self.attributes[button]
+  if type(old) == "string" then
+    old = self.doc:sheets():find(btype, old)
+  end
+  return old
+end
+
 function MODEL:absoluteButton(button)
   -- print("Button:", button)
   if button == "stroke" or button == "fill" then
-    local old = self.doc:sheets():find("color", self.attributes[button])
+    local old = self:findOldValue(button, "color")
     r, g, b = ipeui.getColor(self.ui:win(), "Select " .. button .. " color",
 			     old.r, old.g, old.b)
     if r then
       self:set_absolute(button, { r = r, g = g, b = b });
     end
   elseif button == "pen" then
-    local old = self.doc:sheets():find(button, self.attributes[button])
+    local old = self:findOldValue(button, button)
     local d = self:getDouble("Select pen", "Pen width:", old, 0, 1000)
     if d then
       self:set_absolute(button, d)
     end
   elseif button == "textsize" then
-    local old = self.doc:sheets():find(button, self.attributes[button])
+    local old = self:findOldValue(button, button)
     local d = self:getDouble("Select text size", "Text size in points:",
 			     10, 2, 1000)
     if d then
       self:set_absolute(button, d)
+      self:autoRunLatex()
     end
   elseif button == "symbolsize" then
-    local old = self.doc:sheets():find(button, self.attributes[button])
+    local old = self:findOldValue(button, button)
     local d = self:getDouble("Select symbol size", "Symbol size:",
 			     old, 0, 1000)
     if d then
@@ -231,6 +271,10 @@ function MODEL:action_mark_page(m)
   self:register(t)
 end
 
+function MODEL:action_stop()
+  self.ui:finishTool()
+end
+
 ----------------------------------------------------------------------
 
 function MODEL:showPathStylePopup(v)
@@ -280,6 +324,10 @@ function MODEL:showLayerBoxPopup(v, layer)
   local p = self:page()
   local m = ipeui.Menu(self.ui:win())
   local active = p:active(self.vno)
+  m:add("visible", "Make " .. layer .. " visible from view "
+	.. self.vno .. " onwards")
+  m:add("invisible", "Make " .. layer .. " invisible from view "
+	.. self.vno .. " onwards")
   m:add("rename", "Rename " .. layer)
   m:add("delete", "Delete " .. layer)
   if layer ~= active then
@@ -317,6 +365,39 @@ end
 
 ----------------------------------------------------------------------
 
+function MODEL:layeraction_make_visible(layer, arg, vis)
+  local p = self:page()
+  local label
+  if vis then
+    label = "make layer " .. layer .. " visible from view "
+      .. self.vno .. " onwards"
+  else
+    label = "make layer " .. layer .. " invisible from view "
+      .. self.vno .. " onwards"
+  end
+  local t = { label=label,
+	      pno=self.pno,
+	      vno=self.vno,
+	      original=p:clone(),
+	      layer=layer,
+	      undo=revertOriginal
+	    }
+  t.redo = function (t, doc)
+	     for j = t.vno, doc[t.pno]:countViews() do
+	       doc[t.pno]:setVisible(j, t.layer, vis)
+	     end
+	   end
+  self:register(t)
+end
+
+function MODEL:layeraction_visible(layer, arg)
+  self:layeraction_make_visible(layer, arg, true)
+end
+
+function MODEL:layeraction_invisible(layer, arg)
+  self:layeraction_make_visible(layer, arg, false)
+end
+
 function MODEL:layeraction_select(layer, arg)
   local p = self:page()
   local t = { label="set visibility of layer " .. layer,
@@ -385,9 +466,11 @@ end
 function MODEL:layeraction_move(layer, arg)
   local p = self:page()
   local target = 1
-  if arg ~= "_top_" then target = indexOf(arg, p:layers()) end
   local current = indexOf(layer, p:layers())
-  if target < current then target = target + 1 end
+  if arg ~= "_top_" then
+    target = indexOf(arg, p:layers())
+    if target < current then target = target + 1 end
+  end
   local t = { label="move layer " .. layer,
 	      pno=self.pno,
 	      vno=self.vno,
@@ -426,7 +509,8 @@ end
 
 function MODEL:layeraction_rename(layer)
   local p = self:page()
-  local name = self:getString("Enter new name for layer '" .. layer .. "'")
+  local name = self:getString("Enter new name for layer '" .. layer .. "'",
+			      nil, layer)
   if not name then return end
   name = string.gsub(name, "%s+", "_")
   if indexOf(name, p:layers()) then
@@ -537,8 +621,12 @@ end
 
 ----------------------------------------------------------------------
 
+function action_new_window()
+  MODEL.new()
+end
+
 function MODEL:action_new_window()
-  MODEL:new()
+  MODEL.new()
 end
 
 function MODEL:action_run_latex()
@@ -549,10 +637,6 @@ function MODEL:action_close()
   self.ui:close()
 end
 
-function MODEL:action_quit()
-  ipeui.closeAllWindows()
-end
-
 function MODEL:action_auto_latex()
   self.auto_latex = self.ui:actionState("auto_latex")
   print("Auto is", self.auto_latex)
@@ -565,10 +649,16 @@ function MODEL:action_new()
   self:newDocument()
 end
 
+function action_open()
+  local s, f = ipeui.fileDialog(nil, "open", "Open file", filter_ipe)
+  if s then
+    MODEL.new(nil, s)
+  end
+end
+
 function MODEL:action_open()
   if not self:checkModified() then return end
-  local s, f = ipeui.fileDialog(self.ui:win(), "open", "Open file",
-	  "Ipe files (*.ipe *.pdf *.eps *.xml);;All files (*.*)")
+  local s, f = ipeui.fileDialog(self.ui:win(), "open", "Open file", filter_ipe)
   if s then
     self:loadDocument(s)
   else
@@ -586,23 +676,32 @@ end
 
 function MODEL:action_save_as()
   local dir
-  if self.file_name then dir = self.file_name:match("^(.+)/[^/]+") end
+  if self.file_name then dir = self.file_name:match(prefs.dir_pattern) end
   if not dir then dir = prefs.save_as_directory end
+  local name
+  local filter
+  if self.file_name then
+    name = self.file_name
+    local fmt = formatFromFileName(self.file_name)
+    if fmt then
+      filter = indexOf(fmt, { "xml", "pdf" })
+      name = self.file_name:sub(1,-5)
+    end
+  end
   local s, f =
     ipeui.fileDialog(self.ui:win(), "save", "Save file as",
-		     "XML (*.ipe *.xml);;PDF (*.pdf);;EPS (*.eps)",
-		     dir, self.dialog_filter)
-  local fmap = { XML=".ipe", PDF=".pdf", EPS=".eps" }
+		     filter_save, dir, name, filter)
+  local fmap = { ".ipe", ".pdf" }
   if s then
-    self.dialog_filter = f
     if not formatFromFileName(s) then
-      s = s .. fmap[f:sub(1,3)]
+      s = s .. fmap[f]
     end
-    if ipe.fileExists(s) then
-      local b = ipeui.messageBox(self.ui:win(), "question",
-				 "File already exists!",
-				 "Do you wish to overwrite?<br>" .. s,
-				 "okcancel")
+    -- Cocoa handles the overwrite dialog itself
+    if config.toolkit ~= "cocoa" and ipe.fileExists(s) then
+      local b = messageBox(self.ui:win(), "question",
+			   "File already exists!",
+			   "Do you wish to overwrite '" .. s .. "' ?",
+			   "okcancel")
       if b ~= 1 then
 	self.ui:explain("File not saved")
 	return
@@ -612,18 +711,136 @@ function MODEL:action_save_as()
   end
 end
 
+function MODEL:compute_rect(w, h, res)
+  if res == V(0,0) then res = V(72.0, 72.0) end
+
+  local frame = self.doc:sheets():find("layout").framesize
+
+  local dx = (w * 72.0) / res.x
+  local dy = (h * 72.0) / res.y
+
+  local xfactor = 1.0
+  if (dx > frame.x) then
+    xfactor = frame.x / dx
+  end
+  local  yfactor = 1.0
+  if (dy > frame.y) then
+    yfactor = frame.y / dy
+  end
+  local factor = yfactor
+  if (xfactor < yfactor) then factor = xfactor end
+  dx = dx * factor
+  dy = dy * factor
+
+  local v = 0.5 * V(frame.x - dx, frame.y - dy)
+  local rect = ipe.Rect()
+  rect:add(v)
+  rect:add(v + V(dx, dy))
+  return rect
+end
+
+function MODEL:action_export_png()
+  local dir
+  if self.file_name then dir = self.file_name:match(prefs.dir_pattern) end
+  if not dir then dir = prefs.save_as_directory end
+  local s, f =
+    ipeui.fileDialog(self.ui:win(), "save", "Export to PNG", filter_png, dir)
+  if s then
+    if config.toolkit ~= "cocoa" and ipe.fileExists(s) then
+      local b = messageBox(self.ui:win(), "question",
+			   "File already exists!",
+			   "Do you wish to overwrite?\n\n" .. s,
+			   "okcancel")
+      if b ~= 1 then
+	self.ui:explain("Document not exported")
+	return
+      end
+    end
+    self.ui:renderPage(self.doc, self.pno, self.vno,
+		       "png", s, self.ui:zoom(),
+		       true, false) -- transparent, nocrop
+  end
+end
+
+function MODEL:action_export_eps()
+  local dir
+  if self.file_name then dir = self.file_name:match(prefs.dir_pattern) end
+  if not dir then dir = prefs.save_as_directory end
+  local s, f =
+    ipeui.fileDialog(self.ui:win(), "save", "Export to EPS", filter_eps, dir)
+  if s then
+    if config.toolkit ~= "cocoa" and ipe.fileExists(s) then
+      local b = messageBox(self.ui:win(), "question",
+			   "File already exists!",
+			   "Do you wish to overwrite?\n\n" .. s,
+			   "okcancel")
+      if b ~= 1 then
+	self.ui:explain("Document not exported")
+	return
+      end
+    end
+    self.ui:renderPage(self.doc, self.pno, self.vno,
+		       "eps", s, 1.0, false, false)
+  end
+end
+
+function MODEL:action_insert_image()
+  if not self.insert_image_dir and self.file_name then
+    self.insert_image_dir = self.file_name:match(prefs.dir_pattern)
+  end
+  local s, f = ipeui.fileDialog(self.ui:win(), "open", "Insert image",
+				filter_images, self.insert_image_dir)
+  if s then
+    self.insert_image_dir = s:match(prefs.dir_pattern)
+    local format = "png"
+    if s:sub(-4):lower() == ".jpg" or s:sub(-5):lower() == ".jpeg"
+    then format = "jpeg" end
+    local bitmap, res = ipe.readImage(s, format)
+    if not bitmap then
+      messageBox(self.ui:win(), "warning", "Could not read image",
+		 res, "ok")
+    else
+      local info = bitmap:info()
+      local r = self:compute_rect(info.width, info.height, res)
+      local obj = ipe.Image(r, bitmap)
+      self:creation("insert image", obj)
+    end
+  else
+    self.ui:explain("Insert image canceled")
+  end
+end
+
 ----------------------------------------------------------------------
 
-function MODEL:action_manual()
-  local url = "file:///" .. config.docdir .. "/manual.html"
+function action_manual()
+  local url = config.docdir .. "/manual.html"
   if ipeui.startBrowser then
-    ipeui.startBrowser(url)
+    if not ipeui.startBrowser("file:///" ..  url) then
+      messageBox(nil, "warning", "Manual not available!", url)
+    end
   else
     os.execute(prefs.browser:format(url))
   end
 end
 
-function MODEL:action_show_configuration()
+function MODEL:action_manual()
+  action_manual()
+end
+
+function action_show_libraries(win)
+  local s = "Ipe relies on the following fine pieces of software:\n"
+  s = s .. "\n * Pdftex, Xetex, or Luatex"
+  s = s .. "\n * " .. _VERSION .. " (" .. math.floor(collectgarbage("count")) .. " kB used)"
+  s = s .. "\n * The font rendering library " .. config.freetype_version
+  s = s .. "\n * The rendering library Cairo " .. config.cairo_version
+  if config.qt_version then
+    s = s .. "\n * The GUI toolkit Qt " .. config.qt_version
+  end
+  s = s .. "\n * The compression library zlib " .. config.zlib_version
+  messageBox(win, "information", "Ipe libraries", s)
+end
+
+function action_show_configuration(win)
   local s = ""
   s = s .. " * Lua code: " .. package.path
   s = s .. "\n * Style directories:\n  - " ..
@@ -631,92 +848,241 @@ function MODEL:action_show_configuration()
   s = s .. "\n * Styles for new documents: " ..
     table.concat(prefs.styles, ", ")
   s = s .. "\n * Autosave file: " .. prefs.autosave_filename
+  s = s .. "\n * Save-as directory: " .. prefs.save_as_directory
   s = s .. "\n * Documentation: " .. config.docdir
   s = s .. "\n * Ipelets:\n   - " .. table.concat(config.ipeletDirs, "\n   - ")
+  s = s .. "\n * Latex program path: " .. config.latexpath
   s = s .. "\n * Latex directory: " .. config.latexdir
-  s = s .. "\n * Fontmap: " .. config.fontmap
   s = s .. "\n * Icons: " .. config.icons
-  ipeui.messageBox(self.ui:win(), "information", "Ipe configuration", s)
+  s = s .. "\n * External editor: " .. (prefs.external_editor or "none")
+  messageBox(win, "information", "Ipe configuration", s)
+end
+
+function MODEL:action_show_configuration()
+  action_show_configuration(self.ui:win())
+end
+
+function MODEL:action_show_libraries()
+  action_show_libraries(self.ui:win())
+end
+
+function action_about_ipelets(win)
+  local s = ""
+  for i,v in ipairs(ipelets) do
+    if v.label and v.about then
+      s = s .. "### " .. v.label .. " ###\n\n" .. v.about .. "\n"
+    end
+  end
+  local d = ipeui.Dialog(win, "Ipe: About the ipelets")
+  d:add("text", "text", { read_only=true }, 1, 1)
+  d:set("text", s)
+  d:addButton("ok", "Ok", "accept")
+  d:execute(prefs.latexlog_size)
 end
 
 function MODEL:action_about_ipelets()
-  local s = ""
-  for _,v in ipairs(ipelets) do
-    s = s .. " * " .. v.label .. "\n   " .. v.about .. "\n"
-  end
-  ipeui.messageBox(self.ui:win(), "information", "About the ipelets", s)
+  action_about_ipelets(self.ui:win())
 end
 
 function MODEL:action_keyboard()
   local s = prefs.keyboard
   if s then
-    os.execute(s)
+    if config.platform == "win" then
+      ipeui.startBrowser(s)
+    else
+      os.execute(s)
+    end
   else
     self:warning("No onscreen keyboard defined.",
 		 "Edit preferences to define an onscreen keyboard.")
   end
 end
 
+function MODEL:action_developer_reload_ipelets()
+  load_ipelets()
+  self.ui:explain("Ipelets reloaded")
+end
+
+function MODEL:action_developer_list_shortcuts()
+  local keys = {}
+  for n,k in pairs(shortcuts) do
+    if type(k) == "table" then
+      for _,k1 in ipairs(k) do
+	keys[#keys+1] = k1
+      end
+    else
+      keys[#keys+1] = k
+    end
+  end
+  table.sort(keys)
+  printTable(keys)
+  local dupl = 0
+  local prev = nil
+  for _, k in ipairs(keys) do
+    if k == prev then
+      print("Duplicate shortcut assignment: ", k)
+      dupl = dupl + 1
+    end
+    prev = k
+  end
+  print(dupl, " duplicate shortcuts found")
+end
+
 ----------------------------------------------------------------------
 
-function MODEL:action_set_origin_snap()
+function MODEL:setOrigin(snapping)
   self.snap.origin = self.ui:simpleSnapPos()
-  self.snap.snapangle = true
   self.snap.with_axes = true
-  self.ui:setActionState("snapangle", true)
-  self.ui:setSnap(self.snap)
+  self.ui:setActionState("show_axes", self.snap.with_axes)
+  if snapping then
+    self.snap.snapangle = true
+    self.ui:setActionState("snapangle", true)
+  end
+  self:setSnap()
   self.ui:setFifiVisible(true)
   self.ui:update()
 end
 
-function MODEL:action_hide_axes()
-  self.snap.with_axes = false
-  self.snap.snapangle = false
-  self.ui:setActionState("snapangle", false)
-  self.ui:setSnap(self.snap)
+function MODEL:action_set_origin()
+  self:setOrigin(false)
+end
+
+function MODEL:action_set_origin_snap()
+  self:setOrigin(true)
+end
+
+function MODEL:action_show_axes()
+  self.snap.with_axes = self.ui:actionState("show_axes");
+  if not self.snap.with_axes then
+    self.snap.snapangle = false
+    self.ui:setActionState("snapangle", false)
+  end
+  self:setSnap()
   self.ui:update()
 end
 
 function MODEL:action_set_direction()
   self.snap.with_axes = true;
+  self.ui:setActionState("show_axes", self.snap.with_axes)
   self.snap.orientation = (self.ui:simpleSnapPos() - self.snap.origin):angle()
   self.ui:setSnap(self.snap)
   self.ui:update()
 end
 
+local function find_arc_under_point_in_object(obj, m, pos)
+  if obj:type() == "group" then
+    local m1 = m * obj:matrix()
+    for i = 1,obj:count() do
+      local t = obj:elementType(i)
+      if t == "path" or t == "group" then
+	local obj1 = obj:element(i)
+	local r = ipe.Rect()
+	obj1:addToBBox(r, m1)
+	if r:contains(pos) then
+	  local arc = find_arc_under_point_in_object(obj1, m1, pos)
+	  if arc then return arc end
+	end
+      end
+    end
+  elseif obj:type() == "path" then
+    local m1 = m * obj:matrix()
+    local shape = obj:shape()
+    for _, sp in ipairs(shape) do
+      if sp.type == "ellipse" then
+	local m2 = m1 * sp[1]
+	local post = m2:inverse() * pos
+	if math.abs(post:sqLen() - 1.0) < 1e-3 then
+	  return m2
+	end
+      elseif sp.type == "curve" then
+	for i = 1,#sp do
+	  if sp[i].type == "arc" then
+	    local m2 = m1 * sp[i].arc:matrix()
+	    local post = m2:inverse() * pos
+	    if math.abs(post:sqLen() - 1.0) < 1e-3 then
+	      return m2
+	    end
+	  end
+	end
+      end
+    end
+  end
+end
+
+local function find_arc_under_point(p, pos)
+  local m = ipe.Matrix()
+  for i, obj, sel, layer in p:objects() do
+    if p:bbox(i):contains(pos) then
+      local arc = find_arc_under_point_in_object(obj, m, pos)
+      if arc then return arc end
+    end
+  end
+end
+
+function MODEL:action_set_tangent_direction()
+  local p = self:page()
+  local pos, dir = p:findEdge(self.vno, self.ui:unsnappedPos())
+  if not pos then
+    self.ui:explain("Mouse is not on an edge")
+    return false
+  end
+  local arc = find_arc_under_point(p, pos)
+  if not arc then
+    self.ui:explain("Mouse is not on a circular arc")
+    return false
+  end
+  local p = arc:inverse() * self.snap.origin
+  local d = p:len()
+  if d < 1.0 then
+    self.ui:explain("Origin is inside the ellipse")
+    return false
+  end
+  local r = math.sqrt(d * d - 1.0)
+  local c = ipe.Arc(ipe.Matrix(r, 0, 0, r, p.x, p.y))
+  local qs = c:intersect(ipe.Arc(ipe.Matrix()))
+  if #qs ~= 2 then return false end
+  local pos1 = arc:inverse() * pos
+  local q = qs[1]
+  if (pos1 - qs[2]):sqLen() < (pos1 - q):sqLen() then q = qs[2] end
+  self.snap.with_axes = true;
+  self.ui:setActionState("show_axes", self.snap.with_axes)
+  self.snap.orientation = (arc * q - self.snap.origin):angle()
+  self.ui:setSnap(self.snap)
+  self.ui:update()
+  return true
+end
+
 function MODEL:action_reset_direction()
   self.snap.orientation = 0
-  self.ui:setSnap(self.snap)
+  self:setSnap()
   self.ui:update()
 end
 
-function MODEL:setLine()
-  local origin, dir = self:page():findEdge(self.ui:simpleSnapPos())
+function MODEL:setLine(snapping)
+  local origin, dir = self:page():findEdge(self.vno, self.ui:unsnappedPos())
   if not origin then
     self.ui:explain("Mouse is not on an edge")
-    return false
   else
     self.snap.with_axes = true;
+    self.ui:setActionState("show_axes", self.snap.with_axes)
     self.snap.orientation = dir
     self.snap.origin = origin
     self.ui:update()
-    return true
+    if snapping then
+      self.snap.snapangle = true
+      self.ui:setFifiVisible(true)
+      self.ui:setActionState("snapangle", true)
+    end
+    self:setSnap()
   end
 end
 
 function MODEL:action_set_line()
-  if self:setLine() then
-    self.ui:setSnap(self.snap)
-  end
+  self:setLine(false)
 end
 
 function MODEL:action_set_line_snap()
-  if self:setLine() then
-    self.snap.snapangle = true
-    self.ui:setFifiVisible(true)
-    self.ui:setActionState("snapangle", true)
-    self.ui:setSnap(self.snap)
-  end
+  self:setLine(true)
 end
 
 ----------------------------------------------------------------------
@@ -796,7 +1162,12 @@ end
 
 function MODEL:action_grid_visible()
   self.snap.grid_visible = self.ui:actionState("grid_visible");
-  self.ui:setSnap(self.snap)
+  self:setSnap()
+  self.ui:update()
+end
+
+function MODEL:action_pretty_display()
+  self.ui:setPretty(self.ui:actionState("pretty_display"));
   self.ui:update()
 end
 
@@ -814,31 +1185,32 @@ function MODEL:action_zoom_out()
   self.ui:update()
 end
 
-function MODEL:wheel_zoom(delta)
-  local origin = self.ui:unsnappedPos()
+-- kind is usually 0 except:
+-- kind == 1 on OSX when using a trackpad,
+-- and kind == 2 when holding control
+function MODEL:wheel_zoom(xdelta, ydelta, kind)
+  -- print("wheel_zoom", xdelta, ydelta, kind)
   local zoom = self.ui:zoom()
-  local offset = zoom * (self.ui:pan() - origin)
-  local nzoom
-  if delta > 0 then
-      nzoom = prefs.wheel_zoom_factor * zoom
+  local opts = prefs.scroll
+  if kind == 1 then opts = prefs.trackpad_scroll end
+  if kind ~= 2 and opts.pan then
+    local p = self.ui:pan()
+    local ppd = opts.pixel_per_degree / zoom
+    p = p + V(opts.direction.x * xdelta * ppd, opts.direction.y * ydelta * ppd)
+    self.ui:setPan(p)
   else
-    nzoom = zoom / prefs.wheel_zoom_factor
+    local origin = self.ui:unsnappedPos()
+    local offset = zoom * (self.ui:pan() - origin)
+    local base = opts.zoom_per_degree
+    local nzoom = zoom * base ^ ydelta
+    if nzoom > prefs.max_zoom then nzoom = prefs.max_zoom end
+    if nzoom < prefs.min_zoom then nzoom = prefs.min_zoom end
+    self.ui:setZoom(nzoom)
+    self.ui:setPan(origin + (1/nzoom) * offset)
   end
-  if nzoom > prefs.max_zoom then nzoom = prefs.max_zoom end
-  if nzoom < prefs.min_zoom then nzoom = prefs.min_zoom end
-  self.ui:setZoom(nzoom)
-  self.ui:setPan(origin + (1/nzoom) * offset)
   self.ui:update()
 end
 
-function MODEL:action_wheel_zoom_out()
-  self:wheel_zoom(-1)
-end
-
-function MODEL:action_wheel_zoom_in()
-  self:wheel_zoom(1)
-end
-
 -- Change resolution to 72 dpi and maximize interesting visible area.
 -- As suggested by Rene:
 
@@ -897,7 +1269,7 @@ end
 
 function MODEL:action_normal_size()
   local p = self:page()
-  self.ui:setZoom(1)
+  self.ui:setZoom(prefs.normal_resolution / 72.0)
   local layout = self.doc:sheets():find("layout")
   local paper = ipe.Rect()
   paper:add(-layout.origin)
@@ -1039,6 +1411,10 @@ end
 
 function MODEL:action_delete_view()
   local p = self:page()
+  if p:countViews() < 2 then
+    self.ui:explain("cannot delete single view of a page")
+    return
+  end
   local t = { label="delete view",
 	      pno = self.pno,
 	      vno0 = self.vno,
@@ -1064,9 +1440,9 @@ function MODEL:action_edit_effects()
   local d = ipeui.Dialog(self.ui:win(), "Edit view effect")
   local label = "Choose effect for view " .. self.vno
   d:add("label", "label", { label=label }, 1, 1)
-  d:add("combo", "combo", list, 2, 1)
-  d:addButton("cancel", "&Cancel", "reject")
+  d:add("combo", "combo", list, 0, 1)
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:set("combo", indexOf(effect, list))
   if not d:execute() then return end
   local final = list[d:get("combo")]
@@ -1111,7 +1487,7 @@ end
 
 function MODEL:action_copy_page()
   local data = self:page():xml("ipepage")
-  ipeui.setClipboard(data)
+  self.ui:setClipboard(data)
   self.ui:explain("page copied to clipboard")
 end
 
@@ -1140,7 +1516,7 @@ function MODEL:action_cut_page()
   self:registerOnly(t)
   self.ui:explain(t.label .. " and copied to clipboard")
   local data = t.original:xml("ipepage")
-  ipeui.setClipboard(data)
+  self.ui:setClipboard(data)
 end
 
 function MODEL:action_new_page()
@@ -1162,7 +1538,7 @@ function MODEL:action_new_page()
 end
 
 function MODEL:action_paste_page()
-  local data = ipeui.clipboard()
+  local data = self.ui:clipboard()
   if data:sub(1,9) ~= "<ipepage>" then
     self:warning("No Ipe page to paste")
     return
@@ -1225,22 +1601,22 @@ end
 function MODEL:action_edit_title()
   local d = ipeui.Dialog(self.ui:win(), "Ipe: Edit page title and sections")
   d:add("label1", "label", { label="Page title"}, 1, 1, 1, 4)
-  d:add("title", "text", {}, 2, 1, 1, 4)
-  d:add("label2", "label", { label="Sections"}, 3, 1, 1, 4)
-  d:add("label3", "label", { label="Section:" }, 4, 1)
+  d:add("title", "text", {}, 0, 1, 1, 4)
+  d:add("label2", "label", { label="Sections"}, 0, 1, 1, 4)
+  d:add("label3", "label", { label="Section:" }, 0, 1)
   d:add("tsection", "checkbox",
 	{ label="Use &title",
 	  action=function (d) update_dialog(d, "section") end
-	} , 4, 2)
-  d:add("section", "input", {}, 5, 2, 1, 3)
-  d:add("label4", "label", { label="Subsection:" }, 6, 1)
+	} , -1, 2)
+  d:add("section", "input", {}, 0, 2, 1, 3)
+  d:add("label4", "label", { label="Subsection:" }, 0, 1)
   d:add("tsubsection", "checkbox",
 	{ label="Use titl&e",
 	  action=function (d) update_dialog(d, "subsection") end
-	} , 6, 2)
-  d:add("subsection", "input", {}, 7, 2, 1, 3)
-  d:addButton("cancel", "&Cancel", "reject")
+	} , -1, 2)
+  d:add("subsection", "input", {}, 0, 2, 1, 3)
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("row", 2, 1)
   d:setStretch("column", 2, 1)
   -- setup original values
@@ -1283,15 +1659,17 @@ function MODEL:action_edit_title()
 	     doc[t.pno]:setTitles(t.final)
 	   end
   self:register(t)
+  self:autoRunLatex()
 end
 
 function MODEL:action_edit_notes()
   local d = ipeui.Dialog(self.ui:win(), "Ipe: Edit page notes")
   d:add("label1", "label", { label="Page notes"}, 1, 1)
-  d:add("notes", "text", {}, 2, 1)
-  d:addButton("cancel", "&Cancel", "reject")
+  d:add("notes", "text", {}, 0, 1)
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("row", 2, 1)
+  d:setStretch("column", 1, 1)
   -- setup original values
   local n = self:page():notes()
   d:set("notes", n)
@@ -1346,6 +1724,11 @@ end
 function MODEL:action_page_sorter()
   local arr = self.ui:pageSorter(self.doc)
   if not arr then return end -- canceled
+  if #arr == 0 then
+    self:warning("You cannot delete all pages of the document")
+    return
+  end
+  printTable(arr)
   -- compute reverse 'permutation'
   local rev = {}
   for i = 1,#self.doc do rev[i] = 0 end
@@ -1362,9 +1745,21 @@ function MODEL:action_page_sorter()
   t.redo = function (t, doc)
 	     t.deleted = arrange_pages(doc, t.arr, {})
 	   end
+  self.pno = 1
+  self.vno = 1
   self:register(t)
 end
 
+function MODEL:action_toggle_notes()
+  self.ui:showTool("notes", self.ui:actionState("toggle_notes"))
+  self.ui:setNotes(self:page():notes())
+end
+
+function MODEL:action_toggle_bookmarks()
+  self.ui:showTool("bookmarks", self.ui:actionState("toggle_bookmarks"))
+  self:setBookmarks()
+end
+
 ----------------------------------------------------------------------
 
 function MODEL:action_new_layer()
@@ -1410,32 +1805,37 @@ function MODEL:saction_edit_as_xml()
   local prim = self:page():primarySelection()
   local xml = self:page()[prim]:xml()
   local d = ipeui.Dialog(self.ui:win(), "Edit as XML")
-  d:add("xml", "text", { syntax="xml" }, 1, 1)
+  d:add("xml", "text", { syntax="xml", focus=true }, 1, 1)
   addEditorField(d, "xml")
-  d:addButton("cancel", "&Cancel", "reject")
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("row", 1, 1);
   d:set("xml", xml)
-  if not d:execute() then return end
-  local obj = ipe.Object(d:get("xml"))
-  if not obj then self:warning("Cannot parse XML") return end
+  if prefs.auto_external_editor then
+    externalEditor(d, "xml")
+  end
+  if ((prefs.auto_external_editor and prefs.editor_closes_dialog)
+    or d:execute(prefs.editor_size)) then
+    local obj = ipe.Object(d:get("xml"))
+    if not obj then self:warning("Cannot parse XML") return end
 
-  local t = { label="edit object in XML",
-	      pno = self.pno,
-	      vno = self.vno,
-	      primary = prim,
-	      original = self:page()[prim]:clone(),
-	      final = obj,
-	    }
-  t.undo = function (t, doc)
-	     doc[t.pno]:replace(t.primary, t.original)
-	   end
-  t.redo = function (t, doc)
-	     doc[t.pno]:replace(t.primary, t.final)
-	   end
-  self:register(t)
+    local t = { label="edit object in XML",
+		pno = self.pno,
+		vno = self.vno,
+		primary = prim,
+		original = self:page()[prim]:clone(),
+		final = obj,
+	      }
+    t.undo = function (t, doc)
+	       doc[t.pno]:replace(t.primary, t.original)
+	     end
+    t.redo = function (t, doc)
+	       doc[t.pno]:replace(t.primary, t.final)
+	     end
+    self:register(t)
+    if obj:type() == "text" then self:autoRunLatex() end
+  end
 end
-
 ----------------------------------------------------------------------
 
 function MODEL:saction_group()
@@ -1701,13 +2101,13 @@ end
 
 function MODEL:action_copy()
   local data = self:selectionWithCursor()
-  ipeui.setClipboard(data)
+  self.ui:setClipboard(data)
   self.ui:explain("selection copied to clipboard")
 end
 
 function MODEL:action_cut()
   local data = self:selectionWithCursor()
-  ipeui.setClipboard(data)
+  self.ui:setClipboard(data)
   local t = { label="cut selection",
 	      pno = self.pno,
 	      vno = self.vno,
@@ -1725,15 +2125,24 @@ function MODEL:action_cut()
 end
 
 function MODEL:action_paste()
-  local data = ipeui.clipboard()
-  if data:sub(1,13) ~= "<ipeselection" then
-    self:warning("No Ipe selection to paste")
+  local data = self.ui:clipboard(true)  -- allow a bitmap
+  if not data then
+    self:warning("Nothing to paste")
     return
   end
-  local elements = ipe.Object(data)
-  if not elements then
-    self:warning("Could not parse Ipe selection on clipboard")
-    return
+  local elements
+  if type(data) == "string" then
+    if data:sub(1,13) ~= "<ipeselection" then
+      self:warning("No Ipe selection to paste")
+      return
+    end
+    elements = ipe.Object(data)
+    if not elements then
+      self:warning("Could not parse Ipe selection on clipboard")
+      return
+    end
+  else
+    elements = { data }
   end
   local p = self:page()
   local t = { label="paste objects",
@@ -1771,7 +2180,7 @@ end
 
 function MODEL:action_update_style_sheets()
   local dir = nil
-  if self.file_name then dir = string.gsub(self.file_name, "[^/]+$", "") end
+  if self.file_name then dir = string.gsub(self.file_name, "[^/\\]+$", "") end
   if dir == "" then dir = "." end
   local sheets = self.doc:sheets()
   local qlog = ""
@@ -1798,7 +2207,7 @@ function MODEL:action_update_style_sheets()
     end
   end
 
-  ipeui.messageBox(self.ui:win(), "information",
+  messageBox(self.ui:win(), "information",
 		   "Result of updating stylesheets:", qlog)
 
   local t = { label="update style sheets",
@@ -1829,7 +2238,7 @@ local function sheets_add(d, dd)
   local i = d:get("list")
   if not i then i = 1 end
   local s, f = ipeui.fileDialog(dd.model.ui:win(), "open", "Add stylesheet",
-				"Ipe stylesheets (*.isy)")
+				filter_stylesheets)
   if not s then return end
   local sheet, msg = ipe.Sheet(s)
   if not sheet then
@@ -1852,17 +2261,20 @@ local function sheets_edit(d, dd)
   if not i or dd.list[i]:isStandard() then return end
   local data = dd.list[i]:xml(true)
   local fname = os.tmpname()
+  if prefs.editable_textfile then
+    fname = prefs.editable_textfile
+  end
   local f = io.open(fname, "w")
   f:write(data)
   f:close()
   local sheet, msg
   while not sheet do
-    ipeui.WaitDialog(string.format(prefs.external_editor, fname))
+    ipeui.waitDialog(d, string.format(prefs.external_editor, fname))
     sheet, msg = ipe.Sheet(fname)
     if not sheet then
-      local r = ipeui.messageBox(dd.model.ui:win(), "question",
-				 "Cannot load stylesheet - try again?",
-				 msg, "okcancel")
+      local r = messageBox(dd.model.ui:win(), "question",
+			   "Cannot load stylesheet - try again?",
+			   msg, "okcancel")
       if r <= 0 then
 	os.remove(fname)
 	return
@@ -1915,11 +2327,11 @@ local function sheets_save(d, dd)
   local i = d:get("list")
   if not i then return end
   local s, f = ipeui.fileDialog(dd.model.ui:win(), "save", "Save stylesheet",
-				"Ipe stylesheets (*.isy)")
+				filter_stylesheets)
   if s then
     if s:sub(-4) ~= ".isy" then s = s .. ".isy" end
     local data = dd.list[i]:xml(true)
-    local f, msg = io.open(ipeui.local8bit(s), "wb")
+    local f, msg = ipe.openFile(s, "wb")
     if not f then
       dd.model:warning("Cannot save stylesheet", msg)
       return
@@ -1956,8 +2368,8 @@ function MODEL:action_style_sheets()
 	{ label="&Down", action=function (d) sheets_down(d, dd) end }, 6, 4)
   d:add("save", "button",
 	{ label="&Save", action=function (d) sheets_save(d, dd) end }, 7, 4)
-  d:addButton("cancel", "&Cancel", "reject")
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("column", 2, 1)
   if not d:execute() or not dd.modified then return end
   local t = { label="modify style sheets",
@@ -1980,31 +2392,34 @@ end
 ----------------------------------------------------------------------
 
 function MODEL:action_document_properties()
+  local engines = { "default", "pdftex", "xetex", "luatex" }
   local p = self.doc:properties()
   local d = ipeui.Dialog(self.ui:win(), "Ipe document properties")
   d:add("label1", "label", { label="Title"}, 1, 1)
-  d:add("title", "input", {}, 1, 2, 1, 6)
-  d:add("label2", "label", { label="Author"}, 2, 1)
-  d:add("author", "input", {}, 2, 2, 1, 6)
-  d:add("label3", "label", { label="Subject"}, 3, 1)
-  d:add("subject", "input", {}, 3, 2, 1, 6)
-  d:add("label4", "label", { label="Keywords"}, 4, 1)
-  d:add("keywords", "input", {}, 4, 2, 1, 6)
-  d:add("label5", "label", { label="Latex preamble"}, 5, 1)
-  d:add("preamble", "text", {}, 5, 2, 2, 6)
-  d:add("label6", "label", { label="Page mode"}, 7, 1)
-  d:add("fullscreen", "checkbox", { label="&Full screen"}, 7, 2)
-  d:add("numberpages", "checkbox", { label="&Number pages"}, 7, 3)
-  d:add("label7", "label", { label="Created"}, 8, 1)
-  d:add("created", "label",  { label="" }, 8, 2)
-  d:add("label8", "label", { label="Modified"}, 9, 1)
-  d:add("modified", "label",  { label="" }, 9, 2)
-  d:add("label9", "label", { label="Creator" }, 10, 1)
-  d:add("creator", "label",  { label="" }, 10, 2)
-  d:addButton("cancel", "&Cancel", "reject")
+  d:add("title", "input", {}, -1, 2, 1, 6)
+  d:add("label2", "label", { label="Author"}, 0, 1)
+  d:add("author", "input", {}, -1, 2, 1, 6)
+  d:add("label3", "label", { label="Subject"}, 0, 1)
+  d:add("subject", "input", {}, -1, 2, 1, 6)
+  d:add("label4", "label", { label="Keywords"}, 0, 1)
+  d:add("keywords", "input", {}, -1, 2, 1, 6)
+  d:add("label5", "label", { label="Latex engine"}, 0, 1)
+  d:add("tex", "combo", engines, -1, 2, 1, 6)
+  d:add("label6", "label", { label="Latex preamble"}, 0, 1)
+  d:add("preamble", "text", {}, -1, 2, 2, 6)
+  d:add("label7", "label", { label="Page mode"}, 8, 1)
+  d:add("fullscreen", "checkbox", { label="&Full screen"}, -1, 2)
+  d:add("numberpages", "checkbox", { label="&Number pages"}, -1, 3)
+  d:add("label8", "label", { label="Created"}, 0, 1)
+  d:add("created", "label",  { label="" }, -1, 2)
+  d:add("label9", "label", { label="Modified"}, 0, 1)
+  d:add("modified", "label",  { label="" }, -1, 2)
+  d:add("label10", "label", { label="Creator" }, 0, 1)
+  d:add("creator", "label",  { label="" }, -1, 2)
   d:addButton("ok", "&Ok", "accept")
+  d:addButton("cancel", "&Cancel", "reject")
   d:setStretch("column", 5, 1)
-  d:setStretch("row", 6, 1)
+  d:setStretch("row", 7, 1)
   for n in pairs(p) do d:set(n, p[n]) end
   if not d:execute() then return end
 
@@ -2015,6 +2430,8 @@ function MODEL:action_document_properties()
   for n in pairs(p) do
     if n == "creator" or n == "created" or n == "modified" then
       t.final[n] = p[n]
+    elseif n == "tex" then
+      t.final[n] = engines[d:get(n)]
     else
       t.final[n] = d:get(n)
     end
@@ -2026,6 +2443,30 @@ end
 
 ----------------------------------------------------------------------
 
+function MODEL:saction_set_link_action()
+  local p = self:page()
+  local prim = p:primarySelection()
+  local obj = p[prim]
+  assert(obj:type() == "group")
+  local link_action = self:getString("Enter link action", nil, obj:text())
+  if not link_action then return end
+  if link_action:match("^%s*$)") then link_action = "" end
+  local t = { label="set link action",
+	      pno=self.pno,
+	      vno=self.vno,
+	      primary=prim,
+	      original=obj:clone(),
+	      link_action=link_action,
+	    }
+  t.undo = function (t, doc)
+	     doc[t.pno]:replace(t.primary, t.original)
+	   end
+  t.redo = function (t, doc)
+	     doc[t.pno][t.primary]:setText(t.link_action)
+	   end
+  self:register(t)
+end
+
 function MODEL:saction_remove_clipping()
   local p = self:page()
   local prim = p:primarySelection()
@@ -2060,19 +2501,23 @@ end
 
 function MODEL:saction_add_clipping()
   local p = self:page()
-  local prim = p:primarySelection()
-  local obj = p[prim]
-  assert(obj:type() == "group")
   local sel = self:selection()
   if #sel ~= 2 then
-    self:warning("Must have exactly one secondary selection")
+    self:warning("Must have exactly one two selected objects")
     return
   end
-  local sec = (sel[1] ~= prim) and sel[1] or sel[2]
-  if p[sec]:type() ~= "path" then
-    self:warning("Secondary selection is not a path")
+  local prim = sel[1]
+  local sec = sel[2]
+  if p[prim]:type() ~= "group" then
+    prim = sel[2]
+    sec = sel[1]
+  end
+  local obj = p[prim]
+  if obj:type() ~= "group" or p[sec]:type() ~= "path" then
+    self:warning("Must select exactly one group and one path")
     return
   end
+
   local shape = p[sec]:shape()
   transformShape(p[sec]:matrix(), shape)
   transformShape(obj:matrix():inverse(), shape)
