@@ -209,21 +209,20 @@ prefs.max_zoom = 100
 
 -- When pan is true, the gesture moves (pans) the canvas around, and
 -- you can control the speed of panning with pixel_per_degree.
--- When pan is false, only the vertical scroll movement is used to change the zoom level.
+-- The direction of the movement is determined by direction.x and direction.y (1 or -1).
+-- When pan is false, the vertical scroll movement is used to change the zoom level.
 -- You can control the zoom speed using zoom_per_degree.
--- The direction of the vertical movement is determined by vertical_sign (1 or -1).
+-- If you want to reverse the direction of zooming, change
+-- zoom_per_degree to something smaller than one, e.g. 0.9827
 
 prefs.scroll = { pan = true,
 	         pixel_per_degree = 2,
 		 zoom_per_degree = 1.0176,
-		 vertical_sign = 1 }
+		 direction = { x=1, y=1 } }
 prefs.trackpad_scroll = { pan = true,
 	              	  pixel_per_degree = 2,
 		 	  zoom_per_degree = 1.0176,
-		 	  vertical_sign = 1 }
-if config.platform == "unix" then
-  prefs.scroll.vertical_sign = -1
-end
+			  direction = { x=1, y=1 } }
 
 ----------------------------------------------------------------------
 
