@@ -4,7 +4,7 @@
 --[[
 
     This file is part of the extensible drawing editor Ipe.
-    Copyright (C) 1993-2012  Otfried Cheong
+    Copyright (c) 1993-2019 Otfried Cheong
 
     Ipe is free software; you can redistribute it and/or modify it
     under the terms of the GNU General Public License as published by
@@ -41,10 +41,10 @@ function MODEL:propertiesPopup()
   local sel = self:selection()
   if #sel > 1 then
     self:multiPopup(sel)
-  elseif #sel == 1 then
-    self:singlePopup(self:page():primarySelection())
-  else
+  elseif #sel == 0 then
     self:emptyPopup()
+  else
+    self:singlePopup(self:page():primarySelection())
   end
 end
 
@@ -139,7 +139,11 @@ function MODEL:setAttributeOfPrimary(prim, prop, value)
 				     t.stroke, t.fill)
 	   end
   self:register(t)
-  if prop == "minipage" or prop == "textsize" or prop == "textstyle" then
+  if prop == "minipage" or prop == "textsize" or
+    prop == "textstyle" or prop == "labelstyle" then
+    self:autoRunLatex()
+  end
+  if p[prim]:type() == "text" and prop == "stroke" then
     self:autoRunLatex()
   end
 end
@@ -168,21 +172,58 @@ end
 
 ----------------------------------------------------------------------
 
+local function group_has_text(obj)
+  for i = 1,obj:count() do
+    if obj:elementType(i) == "text" then return true end
+  end
+  return false
+end
+
+local function decorationToName(i, s)
+  if s == "normal" then
+    return "no decoration"
+  else
+    return s:match("^decoration/(.+)$")
+  end
+end
+
+local function decoration_names(model)
+  local symbols = model.doc:sheets():allNames("symbol")
+  local res = { "normal" }
+  for _, name in ipairs(symbols) do
+    if name:match("^decoration/") then
+      res[#res + 1] = name
+    end
+  end
+  return res
+end
+
 function MODEL:properties_group(obj, m)
   m:add("comment", "Group object")
   local s = string.format("%d elements", obj:count())
   local clip = obj:clip()
+  local link_action = obj:text()
   if clip then
     s = s .. ", with clipping"
   end
   m:add("comment", s)
+  if link_action == "" then
+    m:add("action_set_link_action", "Set link action")
+  else
+    m:add("action_set_link_action", "Link action: " .. link_action)
+  end
+  m:add("decoration", "Decoration", decoration_names(self),
+	decorationToName, obj:get("decoration"))
   self:insertBasic(m, obj)
+  m:add("action_apply_properties", "Apply properties")
   if clip then
     m:add("action_remove_clipping", "Remove clipping")
     m:add("action_extract_clipping", "Extract clipping path")
-  else
-    m:add("action_add_clipping", "Add clipping path")
   end
+  if group_has_text(obj) then
+    m:add("action_edit", "Edit text in group")
+  end
+  m:add("action_edit_group", "Edit group")
   m:add("action_edit_as_xml", "Edit as XML")
   m:add("action_ungroup", "Ungroup")
 end
@@ -268,6 +309,8 @@ function MODEL:properties_path(obj, m)
   if #shape > 1 then
     m:add("action_decompose", "Decompose path")
   end
+  m:add("action_pick_properties", "Pick properties")
+  m:add("action_apply_properties", "Apply properties")
   m:add("action_edit_as_xml", "Edit as XML")
   m:add("action_edit", "Edit path")
 end
@@ -278,6 +321,7 @@ function MODEL:properties_text(obj, m)
   local colors = sheet:allNames("color")
   local textsizeL = sheet:allNames("textsize")
   local textstyleL = sheet:allNames("textstyle")
+  local labelstyleL = sheet:allNames("labelstyle")
   m:add("comment", "Text object")
   self:insertBasic(m, obj)
   local t = minipage and "minipage" or "label"
@@ -292,6 +336,8 @@ function MODEL:properties_text(obj, m)
     local ts = obj:get("textstyle")
     m:add("textstyle", "Style: " .. ts, textstyleL, nil, ts)
   else
+    local ts = obj:get("labelstyle")
+    m:add("labelstyle", "Style: " .. ts, labelstyleL, nil, ts)
     local ha = obj:get("horizontalalignment")
     m:add("horizontalalignment", "Horizontal alignment: " .. ha,
 	  {"left", "right", "hcenter"}, nil, ha)
@@ -305,6 +351,8 @@ function MODEL:properties_text(obj, m)
   if minipage then
     m:add("action_change_width", "Change width")
   end
+  m:add("action_pick_properties", "Pick properties")
+  m:add("action_apply_properties", "Apply properties")
   m:add("action_edit_as_xml", "Edit as XML")
   m:add("action_edit", "Edit text")
 end
@@ -314,9 +362,7 @@ function MODEL:properties_image(obj, m)
   local info = obj:info()
   m:add("comment", string.format("%d x %d pixels",
 				  info.width, info.height))
-  m:add("comment", string.format("%d components with %d bits",
-				  info.components, info.bitsPerComponent))
-  m:add("comment", "Filter " .. info.filter)
+  m:add("comment", "Format " .. info.format)
   self:insertBasic(m, obj)
 end
 
@@ -337,6 +383,8 @@ function MODEL:properties_reference(obj, m)
   m:add("symbolsize", "Size: " .. ss, sizes, nil, ss)
   local pen = obj:get("pen")
   m:add("pen", "Pen: " .. pen, pens, nil, pen)
+  m:add("action_pick_properties", "Pick properties")
+  m:add("action_apply_properties", "Apply properties")
   m:add("action_edit_as_xml", "Edit as XML")
 end
 
@@ -412,6 +460,7 @@ function MODEL:multiAttributes(m, tm)
 
   if tm.reference then m:add("symbolsize", "Size", sizes) end
   if tm.text or tm.path then m:add("opacity", "Opacity", opacity) end
+  m:add("action_apply_properties", "Apply properties")
 end
 
 function MODEL:multiPopup()
@@ -435,6 +484,10 @@ function MODEL:multiPopup()
 
   self:multiAttributes(m, typmap)
 
+  if count == 2 and typcount == 2 and typmap["group"] and typmap["path"] then
+    m:add("action_add_clipping", "Add clipping path")
+  end
+
   if typcount == 1 and ttype == 'path' then
     m:add("action_join", "Join paths")
     m:add("action_compose", "Compose paths")
@@ -466,3 +519,150 @@ end
 
 --------------------------------------------------------------------
 
+function MODEL:saction_pick_properties()
+  local p = self:page()
+  local prim = p:primarySelection()
+  local obj = p[prim]
+  self["pick_properties_" .. obj:type()](self, obj)
+  self.ui:setAttributes(self.doc:sheets(), self.attributes)
+end
+
+function MODEL:pick_properties_group(obj)
+  -- nothing
+end
+
+function MODEL:pick_properties_text(obj)
+  local a = self.attributes
+  a.stroke = obj:get("stroke")
+  a.textsize = obj:get("textsize")
+  if obj:get("minipage") then
+    a.textstyle = obj:get("textstyle")
+  else
+    a.horizontalalignment = obj:get("horizontalalignment")
+  end
+  a.verticalalignment = obj:get("verticalalignment")
+  a.opacity = obj:get("opacity")
+end
+
+function MODEL:pick_properties_path(obj)
+  local a = self.attributes
+  a.pathmode = obj:get("pathmode")
+  if a.pathmode ~= "filled" then
+    a.stroke = obj:get("stroke")
+  end
+  if a.pathmode ~= "stroked" then
+    a.fill = obj:get("fill")
+  end
+  a.pen = obj:get("pen")
+  a.dashstyle = obj:get("dashstyle")
+  a.farrow = obj:get("farrow")
+  a.rarrow = obj:get("rarrow")
+  if a.farrow then
+    a.farrowsize = obj:get("farrowsize")
+    a.farrowshape = obj:get("farrowshape")
+  end
+  if a.rarrow then
+    a.rarrowsize = obj:get("rarrowsize")
+    a.rarrowshape = obj:get("rarrowshape")
+  end
+  a.linecap = obj:get("linecap")
+  a.linejoin = obj:get("linejoin")
+  a.fillrule = obj:get("fillrule")
+  a.tiling = obj:get("tiling")
+  a.gradient = obj:get("gradient")
+  a.opacity = obj:get("opacity")
+end
+
+function MODEL:pick_properties_image(obj)
+  -- nothing
+end
+
+function MODEL:pick_properties_reference(obj)
+  local a = self.attributes
+  a.stroke = obj:get("stroke")
+  a.fill = obj:get("fill")
+  a.pen = obj:get("pen")
+  a.symbolsize = obj:get("symbolsize")
+end
+
+----------------------------------------------------------------------
+
+function MODEL:saction_apply_properties()
+  local a = { }
+  for k, v in pairs(self.attributes) do
+    a[k] = v
+  end
+  local t = { label="apply properties",
+	      pno = self.pno,
+	      vno = self.vno,
+	      attributes=a,
+	      selection=self:selection(),
+	      original=self:page():clone(),
+	      undo=revertOriginal,
+	    }
+  t.redo = function (t, doc)
+	     local p = doc[t.pno]
+	     for i,obj,sel,layer in p:objects() do
+	       if sel then
+		 local ty = obj:type()
+		 if ty == "text" then
+		   apply_properties_text(obj, t.attributes)
+		 elseif ty == "path" then
+		   apply_properties_path(obj, t.attributes)
+		 elseif ty == "reference" then
+		   apply_properties_reference(obj, t.attributes)
+		 end
+	       end
+	     end
+	   end
+  self:register(t)
+end
+
+function apply_properties_text(obj, a)
+  obj:set("stroke", a.stroke)
+  obj:set("textsize", a.textsize)
+  if obj:get("minipage") then
+    obj:set("textstyle", a.textstyle)
+  else
+    obj:set("horizontalalignment", a.horizontalalignment)
+  end
+  obj:set("verticalalignment", a.verticalalignment)
+  obj:set("opacity", a.opacity)
+end
+
+function apply_properties_path(obj, a)
+  obj:set("pathmode", a.pathmode)
+  if a.pathmode ~= "filled" then
+    obj:set("stroke", a.stroke)
+  end
+  if a.pathmode ~= "stroked" then
+    obj:set("fill", a.fill)
+  end
+  obj:set("pen", a.pen)
+  obj:set("dashstyle", a.dashstyle)
+  obj:set("farrow", a.farrow)
+  obj:set("rarrow", a.rarrow)
+  if a.farrow then
+    obj:set("farrowsize", a.farrowsize)
+    obj:set("farrowshape", a.farrowshape)
+  end
+  if a.rarrow then
+    obj:set("rarrowsize", a.rarrowsize)
+    obj:set("rarrowshape", a.rarrowshape)
+  end
+  obj:set("linecap", a.linecap)
+  obj:set("linejoin", a.linejoin)
+  obj:set("fillrule", a.fillrule)
+  obj:set("tiling", a.tiling)
+  obj:set("gradient", a.gradient)
+  obj:set("opacity", a.opacity)
+end
+
+function apply_properties_reference(obj, a)
+  obj:set("stroke", a.stroke)
+  obj:set("fill", a.fill)
+  obj:set("pen", a.pen)
+  obj:set("symbolsize", a.symbolsize)
+end
+
+----------------------------------------------------------------------
